-- Filename:     seq_mult_2.vhd
-- Filetype:     VHDL code (behavior)
-- Date:         14 mar 2017
-- Update:       18 apr 2018
-- Description:  Description of the multiplier
-- Author:       J. op den Brouw
-- State:        demo
-- Error:        -
-- Version:      1.0.1
-- Copyright:    (c)2018, De Haagse Hogeschool

-- This file is part of Sequential Multiplier V2.0

-- The VHDL file contains the VHDL description
-- of the sequential multiplier version 2. It is
-- based on the well known add-and-shift school method
-- for multiplication. The pseudo code is as follows:
--
--     P := 0; A := input-A; B := input-B; ready := 0;
--     while B <> 0 loop
--        if b(0) = 1 then
--           P := P + A;
--        end if;
--        left-shift A; right-shift B;
--     end while;
--     ready := 1;
--
-- The description uses Register A, Register P, Register P,
-- the adder and the controller as one datapad process.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity seq_mult_2 is
    generic (n : integer := 4);
    port (clk : in std_logic;
          areset : in std_logic;
          start : in std_logic;
          dataa : in unsigned(n-1 downto 0);
          datab : in unsigned(n-1 downto 0);
          resultp : out unsigned(2*n-1 downto 0);
          ready : out std_logic
          );
end entity seq_mult_2;

architecture rtl of seq_mult_2 is
constant n_zeros : unsigned(n-1 downto 0) := (others => '0');
signal rega : unsigned(2*n-1 downto 0);
signal regb : unsigned(n-1 downto 0);
signal regp : unsigned(2*n-1 downto 0);
signal shifta, shiftb, loada, loadb, loadp, clearp : std_logic;
signal b0, zero : std_logic;
type state_type is (s0, s1, s2);
signal state : state_type;
begin

    -- The next state logic and register logic of the state machine in one process.
    nsl_reg: process (clk, areset) is
    begin
        if areset = '1' then
            state <= s0;
        elsif rising_edge(clk) then
            case state is
                when s0 => if start = '1' then state <= s1; end if;
                when s1 => if zero = '1' then state <= s2; end if;
                when s2 => if start = '0' then state <= s0; end if;
                when others => null;
            end case;
        end if;
    end process;

    -- The output logic of the state machine in one process.
    ol: process (state, b0) is
    begin
        loada <= '0'; loadb <= '0'; loadp <= '0'; shifta <= '0';
        shiftb <= '0'; clearp <= '0'; ready <= '0';
  
        case state is
            when s0 => loada <= '1'; loadb <= '1'; clearp <= '1';
            when s1 => shifta <= '1'; shiftb <= '1';
                       if b0 = '1' then loadp <= '1'; end if;
            when s2 => ready <= '1';
            when others => null;
        end case;
    end process;

    -- The datapad processes the data for register A, register B and register P.
	 -- Furthermore, the process assigns value to the combinatorial outputs
	 -- zero and b0.
    datapad: process (clk, areset, regb) is
    begin
        if areset = '1' then
            rega <= (others => '0');
            regb <= (others => '0');
            regp <= (others => '0');
        elsif rising_edge(clk) then
            -- Register A.
            if loada = '1' then
                rega <=  n_zeros & dataa;
            elsif shifta = '1' then
                rega <= rega(2*n-2 downto 0) & '0';
            end if;
            -- Register B.
            if loadb = '1' then
                regb <= datab;
            elsif shiftb = '1' then
                regb <= '0' & regb(n-1 downto 1);
            end if;
            -- Register P.
            if clearp = '1' then
                regp <= (others => '0');
            elsif loadp = '1' then
                regp <= regp + rega;
            end if;
        end if;
        -- The combinatorial outputs zero and b0
		  if regb = n_zeros then zero <= '1'; else zero <= '0'; end if;
        b0 <= regb(0);
    end process;

    resultp <= regp;
end architecture rtl;